using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using EnvDTE80;

namespace Ora.API
{
    /// <summary>
    /// Very easy-to-use base class for custom groupings. You only have to
    /// override the Map function, which maps a member to a group.
    /// </summary>
    public abstract class MapGrouping : IGrouping
    {
        private readonly string _name;
        private readonly Bitmap _bitmap;

        /// <summary>
        /// Constructor taking a name for the grouping.
        /// </summary>
        /// <param name="name">The name</param>
        protected MapGrouping(string name)
        {
            _name = name;
        }

        /// <summary>
        /// Constructor taking a name and an icon to use for each group.
        /// </summary>
        /// <param name="name">The name</param>
        /// <param name="bitmap">The icon</param>
        protected MapGrouping(string name, Bitmap bitmap)
            : this(name)
        {
            _bitmap = bitmap;
        }

        /// <summary>
        /// Must be overridden. Examine the member (and optionally the parent)
        /// parameters and return a string indicating which group the member
        /// should go in. The returned string may be null or empty, in which
        /// case the member is not grouped.
        /// </summary>
        /// <param name="member"></param>
        /// <param name="parent"></param>
        /// <returns></returns>
        protected abstract string Map(CodeElement2 member, CodeElement2 parent);

        /// <summary>
        /// Returns the grouping name supplied to the constructor.
        /// </summary>
        public string Name
        {
            get { return _name; }
        }

        /// <summary>
        /// Stores the next grouping in the current chain.
        /// </summary>
        public IAnonymousGrouping NextInChain { get; set; }

        /// <summary>
        /// Calls Map for each member and outputs the results.
        /// </summary>
        /// <param name="node"></param>
        /// <param name="members"></param>
        /// <param name="parent"></param>
        public void Generate(IResultsTreeNode node, IEnumerable<CodeElement2> members, CodeElement2 parent)
        {
            Dictionary<string, List<CodeElement2>> dictionary = new Dictionary<string, List<CodeElement2>>();
            List<CodeElement2> nulls = new List<CodeElement2>();

            foreach (CodeElement2 member in members)
            {
                if (node.ShouldAbort)
                    return;

                string key = Map(member, parent);

                if (string.IsNullOrEmpty(key))
                    nulls.Add(member);
                else
                {
                    List<CodeElement2> list;
                    if (!dictionary.TryGetValue(key, out list))
                    {
                        list = new List<CodeElement2>();
                        dictionary.Add(key, list);
                    }

                    list.Add(member);
                }
            }
            
            NextInChain.Generate(node, nulls, parent);

            List<string> keys = dictionary.Keys.ToList();
            keys.Sort();

            foreach (string key in keys)
            {
                if (node.ShouldAbort)
                    return;

                NextInChain.Generate(node.Add(key, _bitmap), dictionary[key], parent);
            }
        }
    }
}