using System.Collections.Generic;
using EnvDTE;
using EnvDTE80;
using Ora.API;

namespace Ora.Groupings
{
    public class Polymorphism : IGrouping
    {
        public string Name
        {
            get { return "Polymorphism"; }
        }

        public IAnonymousGrouping NextInChain { get; set; }
        
        private static readonly Dictionary<string, List<CodeElement2>> Categories = new Dictionary<string, List<CodeElement2>>();
        
        private static List<CodeElement2> GetCategory(string name)
        {
            List<CodeElement2> result;
            if (Categories.TryGetValue(name, out result))
                return result;

            result = new List<CodeElement2>();
            Categories.Add(name, result);
            return result;
        }

        private static readonly List<CodeElement2> Other = GetCategory("other");
        private static readonly List<CodeElement2> Constructors = GetCategory("constructors");
        private static readonly List<CodeElement2> Operators = GetCategory("operators");
        private static readonly List<CodeElement2> Statics = GetCategory("statics");
        private static readonly List<CodeElement2> Overrides = GetCategory("overrides");
        private static readonly List<CodeElement2> Sealeds = GetCategory("sealeds");
        private static readonly List<CodeElement2> Virtuals = GetCategory("virtuals");
        private static readonly List<CodeElement2> Abstracts = GetCategory("abstracts");
        
        private static bool Classify(CodeElement2 member, vsCMOverrideKind kind)
        {
            switch (kind)
            {
                case vsCMOverrideKind.vsCMOverrideKindOverride:
                    Overrides.Add(member);
                    return true;

                case vsCMOverrideKind.vsCMOverrideKindSealed:
                    Sealeds.Add(member);
                    return true;

                case vsCMOverrideKind.vsCMOverrideKindVirtual:
                    Virtuals.Add(member);
                    return true;

                case vsCMOverrideKind.vsCMOverrideKindAbstract:
                    Abstracts.Add(member);
                    return true;
            }

            return false;
        }

        private static void Classify(CodeElement2 member)
        {
            CodeEvent evt = member as CodeEvent;
            if (evt != null)
            {
                if (Classify(member, evt.OverrideKind))
                    return;
            }
            else
            {
                CodeProperty prop = member as CodeProperty;
                if (prop != null)
                {
                    CodeFunction2 func = (prop.Getter as CodeFunction2) ?? (prop.Setter as CodeFunction2);
                    if ((func != null) && Classify(member, func.OverrideKind))
                        return;
                }
                else
                {
                    CodeFunction2 func = member as CodeFunction2;
                    if (func != null)
                    {
                        switch (func.FunctionKind)
                        {
                            case vsCMFunction.vsCMFunctionConstructor:
                                Constructors.Add(member);
                                return;

                            case vsCMFunction.vsCMFunctionOperator:
                                Operators.Add(member);
                                return;

                            case vsCMFunction.vsCMFunctionShared:
                                Statics.Add(member);
                                return;
                        }

                        if (Classify(member, func.OverrideKind))
                            return;
                    }
                }
            }

            Other.Add(member);
        }

        public void Generate(IResultsTreeNode node, IEnumerable<CodeElement2> members, CodeElement2 parent)
        {
            Other.Clear();
            Constructors.Clear();
            Operators.Clear();
            Statics.Clear();
            Overrides.Clear();
            Sealeds.Clear();
            Virtuals.Clear();
            Abstracts.Clear();

            foreach (CodeElement2 member in members)
                Classify(member);

            foreach (string grouping in new[] { "constructors", "operators", "statics", 
                                    "overrides", "sealeds", "virtuals", "abstracts" })
            {
                if (node.ShouldAbort)
                    return;

                NextInChain.Generate(node.Add(grouping, null), GetCategory(grouping), parent);
            }

            NextInChain.Generate(node, Other, parent);
        }
    }
}