using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using EnvDTE;
using EnvDTE80;

namespace Ora
{
    public static class EnvDTEExtensions
    {
        public static CodeElement2 GetParent(this CodeElement2 elem)
        {
            CodeClass2 cls = elem as CodeClass2;
            if (cls != null)
                return (CodeElement2)cls.Parent;

            CodeStruct2 strct = elem as CodeStruct2;
            if (strct != null)
                return (CodeElement2)strct.Parent;

            CodeInterface2 iface = elem as CodeInterface2;
            if (iface != null)
                return (CodeElement2)iface.Parent;

            return null;
        }

        public static bool IsAncestorOf(this CodeElement2 elem, CodeElement2 other)
        {
            while (other != null)
            {
                if (other == elem)
                    return true;

                other = other.GetParent();
            }

            return false;
        }

        private static int CompareDepth(CodeElement2 a, CodeElement2 b)
        {
            if (a.IsAncestorOf(b))
                return 1;

            if (b.IsAncestorOf(a))
                return -1;

            return 0;
        }

        public static CodeElement2 TryGetFirstElement(this CodeElement2 elem)
        {
            CodeClass2 cls = elem as CodeClass2;
            if (cls != null)
                return elem;

            CodeStruct2 strct = elem as CodeStruct2;
            if (strct != null)
                return elem;

            CodeInterface2 iface = elem as CodeInterface2;
            if (iface != null)
                return elem;

            foreach (CodeElement2 child in elem.Children)
            {
                CodeElement2 found = child.TryGetFirstElement();
                if (found != null)
                    return found;
            }

            return null;
        }

        public static CodeElement2 TryGetFirstElement(this Document document)
        {
            FileCodeModel2 codeModel = document.ProjectItem.FileCodeModel as FileCodeModel2;
            if (codeModel != null)
            {
                try
                {
                    foreach (CodeElement2 elem in codeModel.CodeElements)
                    {
                        CodeElement2 child = elem.TryGetFirstElement();
                        if (child != null)
                            return child;
                    }
                }
                catch (Exception)
                {
                    return null;
                }
            }

            return null;
        }

        public static CodeElement2 GetMemberUnderCursor(this Document document)
        {
            TextDocument textDoc = document.Object(string.Empty) as TextDocument;
            if (textDoc != null)
            {
                VirtualPoint cursor = textDoc.Selection.ActivePoint;

                return cursor.get_CodeElement(vsCMElement.vsCMElementDelegate) as CodeElement2 ??
                       cursor.get_CodeElement(vsCMElement.vsCMElementEnum) as CodeElement2 ??
                       cursor.get_CodeElement(vsCMElement.vsCMElementEvent) as CodeElement2 ??
                       cursor.get_CodeElement(vsCMElement.vsCMElementFunction) as CodeElement2 ??
                       cursor.get_CodeElement(vsCMElement.vsCMElementProperty) as CodeElement2 ??
                       cursor.get_CodeElement(vsCMElement.vsCMElementVariable) as CodeElement2 ??
                       cursor.get_CodeElement(vsCMElement.vsCMElementInterface) as CodeElement2 ??
                       cursor.get_CodeElement(vsCMElement.vsCMElementStruct) as CodeElement2 ??
                       cursor.get_CodeElement(vsCMElement.vsCMElementClass) as CodeElement2;
            }

            return null;
        }

        public static CodeElement2 TryGetCodeElementFromPoint(this Document document, VirtualPoint cursor, vsCMElement scope)
        {
            FileCodeModel2 codeModel = document.ProjectItem.FileCodeModel as FileCodeModel2;
            if (codeModel != null)
            {
                try
                {
                    return codeModel.CodeElementFromPoint(cursor, scope) as CodeElement2;
                }
                catch (Exception)
                {
                    return null;
                }
            }

            return null;
        }

        public static CodeElement2 GetElementUnderCursor(this Document document)
        {
            if (document == null)
                return null;

            TextDocument textDoc = document.Object(string.Empty) as TextDocument;
            if (textDoc != null)
            {
                VirtualPoint cursor = textDoc.Selection.ActivePoint;

                List<CodeElement2> found = new List<CodeElement2>
                   {
                       document.TryGetCodeElementFromPoint(cursor, vsCMElement.vsCMElementClass),
                       document.TryGetCodeElementFromPoint(cursor, vsCMElement.vsCMElementStruct),
                       document.TryGetCodeElementFromPoint(cursor, vsCMElement.vsCMElementInterface),
                   };
                
                found.RemoveAll(e => e == null);
                found.Sort(CompareDepth);

                if (found.Count != 0)
                    return found[0];    
            }

            return document.TryGetFirstElement();
        }

        public static IEnumerable<CodeInterface2> GetInterfaces(this CodeInterface2 i)
        {
            yield return i;

            foreach (CodeInterface2 j in i.Bases)
            {
                foreach (CodeInterface2 k in j.GetInterfaces())
                    yield return k;
            }
        }

        public static IEnumerable<CodeInterface2> GetInterfaces(this CodeElement2 elem)
        {
            CodeElements ifaces = null;

            CodeClass2 cls = elem as CodeClass2;
            if (cls != null)
                ifaces = cls.ImplementedInterfaces;
            else
            {
                CodeStruct2 strct = elem as CodeStruct2;
                if (strct != null)
                    ifaces = strct.ImplementedInterfaces;
                else
                {
                    CodeInterface2 iface = elem as CodeInterface2;
                    if (iface != null)
                        ifaces = iface.Bases;
                }
            }

            if (ifaces == null)
                yield break;
            
            foreach (CodeInterface2 i in ifaces)
            {
                foreach (CodeInterface2 j in i.GetInterfaces())
                    yield return j;
            }
        }

        public static CodeInterface2 GetAssociatedInterface(this CodeElement2 elem, CodeElement2 cls,
                                                            Func<CodeElement2, string> getSignature)
        {
            string sig = getSignature(elem);

            string explicitName = null;
            int dot = elem.Name.LastIndexOf('.');
            if (dot != -1)
            {
                explicitName = elem.Name.Substring(0, dot);

                dot = explicitName.LastIndexOf('.');
                if (dot != -1)
                    explicitName = explicitName.Substring(dot + 1);
            }

            if (explicitName != null)
            {
                foreach (CodeInterface2 interfaceElement in cls.GetInterfaces())
                {
                    if (interfaceElement.Name == explicitName)
                    {
                        foreach (CodeElement2 interfaceMember in interfaceElement.Members)
                        {
                            if (getSignature(interfaceMember) == sig)
                                return interfaceElement;
                        }
                    }
                }
            }

            foreach (CodeInterface2 interfaceElement in cls.GetInterfaces())
            {
                foreach (CodeElement2 interfaceMember in interfaceElement.Members)
                {
                    if (getSignature(interfaceMember) == sig)
                        return interfaceElement;
                }
            }

            return null;
        }

        private static readonly Dictionary<string, string> _typeAliasNames = new[]
            {
                new { Key = "System.String", Value = "string" },
                new { Key = "System.SByte", Value = "sbyte" },
                new { Key = "System.Byte", Value = "byte" },
                new { Key = "System.Char", Value = "char" },
                new { Key = "System.Int16", Value = "short" },
                new { Key = "System.UInt16", Value = "ushort" },
                new { Key = "System.Int32", Value = "int" },
                new { Key = "System.UInt32", Value = "uint" },
                new { Key = "System.Int64", Value = "long" },
                new { Key = "System.UInt64", Value = "ulong" },
                new { Key = "System.Single", Value = "float" },
                new { Key = "System.Double", Value = "double" },
                new { Key = "System.Decimal", Value = "decimal" },
                new { Key = "System.Bool", Value = "bool" }

            }.ToDictionary(i => i.Key, i => i.Value);

        public static string TrimNonGenericType(string name)
        {
            string alias;
            if (_typeAliasNames.TryGetValue(name, out alias))
                return alias;

            int dot = name.LastIndexOf('.');
            if (dot == -1)
                return name;

            return name.Substring(dot + 1);
        }

        public static string TrimType(string name)
        {
            StringBuilder result = new StringBuilder();

            int start = 0;
            for (int n = 0; n < name.Length; n++)
            {
                char c = name[n];

                if (c == '<' || c == '>' || c == ',')
                {
                    result.Append(TrimNonGenericType(name.Substring(start, n - start)));
                    result.Append(c);
                    start = n + 1;
                }
            }

            result.Append(TrimNonGenericType(name.Substring(start)));
            return result.ToString();
        }
    }
}