using System.Collections.Generic;
using System.Linq;
using EnvDTE80;
using Ora.API;

namespace Ora
{
    public static class GroupingExtensions
    {
        /// <summary>
        /// Doesn't use NextInChain, so can act as a general-purpose end-of-chain
        /// </summary>
        private class Terminal : IAnonymousGrouping
        {
            public IAnonymousGrouping NextInChain { get; set; }
            
            public List<T> ToList<T>(IResultsTreeNode node, IEnumerable<T> source)
            {
                List<T> result = new List<T>();

                foreach (T item in source)
                {
                    if (node.ShouldAbort)
                        return new List<T>();

                    result.Add(item);
                }

                return result;
            }

            public void Generate(IResultsTreeNode node, IEnumerable<CodeElement2> members, CodeElement2 parent)
            {
                List<CodeElement2> sorted = ToList(node, members);
                sorted.Sort((a, b) => a.GetMemberKind().GetCaption(a).CompareTo(b.GetMemberKind().GetCaption(b)));

                foreach (CodeElement2 member in sorted)
                {
                    if (node.ShouldAbort)
                        return;

                    node.Add(member);
                }
            }
        }

        /// <summary>
        /// Plugs together several groupings into a chain or pipeline, so the resulting
        /// combination can be used to build a tree in which multiple groupings are applied
        /// in the order specified.
        /// </summary>
        /// <param name="groupings"></param>
        /// <returns></returns>
        public static IAnonymousGrouping Chain(this IEnumerable<IGrouping> groupings)
        {
            IAnonymousGrouping next = new Terminal();

            foreach (IGrouping grouping in groupings.Reverse())
            {
                grouping.NextInChain = next;
                next = grouping;
            }

            return next;
        }

        /// <summary>
        /// Takes a comma delimited list of grouping names and returns an ordered set
        /// containing those named groupings. This is the opposite of GenerateQueryString.
        /// </summary>
        /// <param name="all">The list of all known groupings</param>
        /// <param name="list">The string specifying an ordered subset of groupings</param>
        /// <returns></returns>
        public static IEnumerable<IGrouping> ParseQueryString(this IEnumerable<IGrouping> all, string list)
        {
            return list.Split(',').Select(name => all.SingleOrDefault(query => query.Name == name.Trim()))
                                  .Where(query => query != null);
        }

        /// <summary>
        /// Returns the names of a set of groupings, separated by commas. This is known as
        /// a "query string", and appears in the Ora toolbar and is also used to persist
        /// the configuration to the registry. It can be parsed back into a list of grouping
        /// objects by ParseQueryString.
        /// </summary>
        /// <param name="selected"></param>
        /// <returns></returns>
        public static string GenerateQueryString(this IEnumerable<IGrouping> selected)
        {
            if (!selected.Any())
                return string.Empty;

            return selected.Select(query => query.Name).Aggregate((a, b) => a + ", " + b);
        }
    }
}