﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Windows.Forms;
using EnvDTE;
using EnvDTE80;
using Ora.API;
using Reflect;
using VersionCheck;
using Process=System.Diagnostics.Process;
using Thread=System.Threading.Thread;

namespace Ora
{
    public partial class OraPane : UserControl
    {
        private CodeElement2 _classElem;
        private bool _yieldFocus;

        private readonly List<IGrouping> _groupings = Components.Construct<IGrouping>().ToList();
        private List<IGrouping> _selected;
        private readonly Images _images = new Images();
        private int _updateCounter;
        private int _disableSelectNotify;
        private int _buildAvailable;
        private const int BuildInstalled = 13;

        private readonly LinkLabel _linkLabel = new LinkLabel
            {
                Anchor = AnchorStyles.Top | AnchorStyles.Bottom | AnchorStyles.Left | AnchorStyles.Right,
                TextAlign = System.Drawing.ContentAlignment.MiddleCenter
            };

        public OraPane()
        {
            InitializeComponent();

            timer1.Interval = 500;
            timer1.Enabled = true;

            _selected = _groupings.ParseQueryString(Configuration.Groupings).ToList();

            treeView1.ImageList = _images.List;
            toolStripLabel1.Text = _selected.GenerateQueryString();

            Populate();

            Connect.CodeModelEvents.ElementAdded += CodeModelEventsElementAdded;
            Connect.CodeModelEvents.ElementChanged += CodeModelEventsElementChanged;
            Connect.CodeModelEvents.ElementDeleted += CodeModelEventsElementDeleted;

            _linkLabel.Click += LinkLabelClick;

            new Thread(VersionCheckThread).Start();
        }

        void VersionCheckThread()
        {
            try
            {
                _buildAvailable = Versioning.GetLatestBuildNumber(FeedClient.ReadTitles(
                    "http://www.codeplex.com/ora/Project/ProjectRss.aspx?ProjectRSSFeed=codeplex%3a%2f%2frelease%2fora"));
            }
// ReSharper disable EmptyGeneralCatchClause
            catch
// ReSharper restore EmptyGeneralCatchClause
            {

            }
        }

        void CodeModelEventsElementChanged(CodeElement element, vsCMChangeKind change)
        {
            DelayedUpdate();
        }

        void CodeModelEventsElementDeleted(object parent, CodeElement element)
        {
            DelayedUpdate();
        }

        void CodeModelEventsElementAdded(CodeElement element)
        {
            DelayedUpdate();
        }

        private void DelayedUpdate()
        {
            _updateCounter = 7;
            _classElem = null;
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            if (_buildAvailable > BuildInstalled)
                toolStripLabel1.Text = string.Format("Get build {0}!", _buildAvailable);

            try
            {
                if (Connect.VisualStudioApplication.ActiveDocument != null)
                {
                    if (_yieldFocus)
                    {
                        Connect.VisualStudioApplication.ActiveDocument.ActiveWindow.SetFocus();
                        _yieldFocus = false;
                    }
                    else
                    {
                        CodeElement2 sel = Connect.VisualStudioApplication.ActiveDocument.GetMemberUnderCursor();
                        if (sel != null)
                        {
                            CodeElement2 cur = (treeView1.SelectedNode != null)
                                                   ? (treeView1.SelectedNode.Tag as CodeElement2)
                                                   : null;

                            if (sel != cur)
                            {
                                _disableSelectNotify++;

                                try
                                {
                                    SelectNode(treeView1.Nodes, sel);
                                }
                                finally
                                {
                                    _disableSelectNotify--;
                                }
                            }
                        }
                    }
                }

                if (_updateCounter == 0)
                    TrackCurrentClass();
                else
                    _updateCounter--;
            }
// ReSharper disable EmptyGeneralCatchClause
            catch
// ReSharper restore EmptyGeneralCatchClause
            {
                // General catch to deal with things like the Project properties, which causes
                // Connect.VisualStudioApplication.ActiveDocument to throw an exception when
                // you try to access it!
            }
        }
        
        private bool SelectNode(TreeNodeCollection nodes, CodeElement2 member)
        {
            foreach (TreeNode node in nodes)
            {
                if ((node.Tag as CodeElement2) == member)
                {
                    treeView1.SelectedNode = node;    
                    return true;
                }

                if (SelectNode(node.Nodes, member))
                    return true;
            }

            return false;
        }

        private void TrackCurrentClass()
        {
            CodeElement2 classElem = null;

            try
            {
                Document doc = Connect.VisualStudioApplication.ActiveDocument;
                if (doc != null)
                    classElem = doc.GetElementUnderCursor();
            }
            catch (Exception e)
            {
                LogException(e);
            }

            if (_classElem != classElem)
            {
                _classElem = classElem;
                Populate();
            }
        }

// ReSharper disable UnusedParameter.Local
        private void LogException(Exception e)
// ReSharper restore UnusedParameter.Local
        {
            while (e.InnerException != null)
                e = e.InnerException;

            ShowLink(e.Message + " - click to retry");
        }

        void ShowControl(Control ctrl)
        {
            foreach (Control other in new Control[] { _linkLabel, treeView1 }.Where(c => c != ctrl))
                tableLayoutPanel1.Controls.Remove(other);

            if (!tableLayoutPanel1.Controls.Contains(ctrl))
                tableLayoutPanel1.Controls.Add(ctrl, 1, 1);
        }

        void ShowLink(string message)
        {
            ShowControl(_linkLabel);
            _linkLabel.Text = message;
        }

        void LinkLabelClick(object sender, EventArgs e)
        {
            Populate(() => false);
        }

        private void Populate()
        {
            bool timedOut = false;

            Stopwatch timer = new Stopwatch();
            timer.Start();

            Populate(() =>
                         {
                             if (timedOut)
                                 return true;

                             if (timer.ElapsedMilliseconds > Configuration.RefreshTimeout)
                             {
                                 timedOut = true;
                                 return true;
                             }

                             return false;
                         });

            if (timedOut)
                ShowLink("Too slow - click to get results (consider refactoring your class!)");
        }

        private void Populate(Func<bool> shouldAbort)
        {
            ShowControl(treeView1);

            treeView1.Visible = false;

            treeView1.Nodes.Clear();
            if (_classElem == null)
                treeView1.Nodes.Add("(Nothing to show)");
            else
            {
                try
                {
                    ResultsTreeNode root = new ResultsTreeNode(treeView1.Nodes, _images, shouldAbort);
                    _selected.Chain().Generate(root, _classElem.Children.OfType<CodeElement2>(), _classElem);

                    if (root.Empty())
                        treeView1.Nodes.Add("(No members)");

                    treeView1.ExpandAll();
                }
                catch (Exception x)
                {
                    ShowLink(x.Message);
                }
            }

            treeView1.Visible = true;
        }

        private void treeView1_AfterSelect(object sender, TreeViewEventArgs e)
        {
            if (_disableSelectNotify != 0)
                return;

            TreeNode node = treeView1.SelectedNode;
            if (node != null)
            {
                CodeElement element = node.Tag as CodeElement;
                if (element != null)
                {
                    try
                    {
                        TextPoint position = element.StartPoint;

                        Document doc = Connect.VisualStudioApplication.ActiveDocument;
                        if (doc != null)
                        {
                            TextSelection textSelection = doc.ActiveWindow.Selection as TextSelection;

                            if (textSelection != null)
                            {
                                textSelection.MoveToPoint(position, false);
                                _yieldFocus = true;
                            }
                        }
                    }
                    catch (Exception x)
                    {
                        MessageBox.Show(x.Message);
                    }
                }
            }
        }

        private void toolStripButton1_Click(object sender, EventArgs e)
        {
            ConfigurationDialog dlg = new ConfigurationDialog
                                          {
                                              Groupings = _groupings,
                                              Selected = _selected,
                                              HideMethodUnderscores = Configuration.HideMethodUnderscores,
                                              RefreshTimeout = Configuration.RefreshTimeout
                                          };

            if (dlg.ShowDialog(this) == DialogResult.OK)
            {
                _selected = dlg.Selected;
                toolStripLabel1.Text = _selected.GenerateQueryString();
                Configuration.Groupings = toolStripLabel1.Text;
                Configuration.HideMethodUnderscores = dlg.HideMethodUnderscores;
                Configuration.RefreshTimeout = dlg.RefreshTimeout;
                Populate();
            }
        }

        private void toolStripLabel1_Click(object sender, EventArgs e)
        {
            if (_buildAvailable > BuildInstalled)
            {
                Process.Start("http://www.codeplex.com/ora");
                toolStripLabel1.Text = _selected.GenerateQueryString();
                _buildAvailable = 0;
            }
            else
            {
                toolStripButton1_Click(sender, e);
            }
        }

        private void toolStripButton2_Click(object sender, EventArgs e)
        {
            try
            {
                TextSelection textSelection = Connect.VisualStudioApplication.ActiveDocument.ActiveWindow.Selection as TextSelection;
                if (textSelection != null)
                {
                    TextDocument doc = textSelection.Parent;
                    if (doc != null)
                    {
                        try
                        {
                            doc.DTE.UndoContext.Open("Remove regions", false);

                            const string startOfLine = "^";
                            const string endOfLine = "$";
                            const string whiteSpace = @"[ \t]*";
                            const string hashChar = @"\u0023";
                            const string anything = ".*";

                            const string preprocDirective =
                                startOfLine +
                                whiteSpace +
                                hashChar +
                                whiteSpace +
                                "{0}" +
                                whiteSpace +
                                anything +
                                endOfLine;

                            string beginRegion = string.Format(preprocDirective, "region");
                            string endRegion = string.Format(preprocDirective, "endregion");

                            string regionDirectives = string.Format("({0})|({1})", beginRegion, endRegion);
                                         
                            TextRanges ranges = null;

                            while (doc.ReplacePattern(regionDirectives, " ",
                                               (int)(vsFindOptions.vsFindOptionsRegularExpression) |
                                               (int)(vsFindOptions.vsFindOptionsFromStart), ref ranges)) { }
                        }
                        finally
                        {
                            doc.DTE.UndoContext.Close();
                        }

                        _yieldFocus = true;
                    }
                }
            }
            catch (Exception x)
            {
                MessageBox.Show(x.Message);
            }
        }

        private void tableLayoutPanel1_Paint(object sender, PaintEventArgs e)
        {

        }
    }
}
