﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace Reflect
{
    /// <summary>
    /// The generic version of Components acts as a reusable factory for a set
    /// of "components", defined as all the types in the local assemblies that
    /// implement a specific interface (or base class).
    /// </summary>
    /// <typeparam name="TInterface"></typeparam>
    public sealed class Components<TInterface> where TInterface : class
    {
        private readonly List<Type> _components = new List<Type>();

        public Components()
        {
            _components = Types.All.Where(type => (typeof(TInterface).IsAssignableFrom(type) && 
                                                   type.IsConstructable())).ToList();
        }

        /// <summary>
        /// Returns a set of new instances of all the available components.
        /// </summary>
        /// <returns></returns>
        public IEnumerable<TInterface> Construct()
        {
            return _components.Select(type => type.Construct()).OfType<TInterface>();
        }
    }

    /// <summary>
    /// The non-generic version of Components is used to get a one-off set of 
    /// instances of "components".
    /// </summary>
    public static class Components
    {
        /// <summary>
        /// Returns a newly constructed instance of every type in every local assembly that
        /// implements the specified interface (or base class).
        /// </summary>
        /// <typeparam name="TInterface">The interface (or base class) that components must support.</typeparam>
        /// <returns></returns>
        public static IEnumerable<TInterface> Construct<TInterface>() where TInterface : class
        {
            return new Components<TInterface>().Construct();
        }
    }
}
